package folio;

import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.SAXParser;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class Annotations implements SAXParserHandler
{
    public static BufferedWriter errors;
    private static Regex reAnnotation = new Regex("", "(annotations.*)");
    private static Regex re = new Regex("", "^(f.\\s*\\d+[rv])");
    private ArrayList<String> comments = new ArrayList<String>();
    private Locator locator;
    private String filename;
    private String shortname;
    private String currentElement;
    private int annotationNumber;
    private SAXParser saxParser;
    private StringBuilder buffer = new StringBuilder();
    private String currentFolio = null;
    private ArrayList<String> anupamaComments;

    public Annotations () {
        saxParser = new SAXParser(this);
    }

    public void parse(String filename, ArrayList<String> anupamaComments) throws Exception {
        this.filename = filename;
        this.anupamaComments = anupamaComments;
        annotationNumber = 0;
        currentElement = null;
        currentFolio = "";
        buffer.setLength(0);
        shortname = new File(filename).getName();
        shortname = shortname.substring(0, shortname.indexOf("."));
        try {
            saxParser.parse(filename);
        } catch (Exception ex) {
            reAnnotation.reset(filename);
            boolean displayError = true;
            String g = null;
            if (reAnnotation.find()) {
                g = reAnnotation.getGroup(1);
                if (Errors.missingAnnotations.get(g) != null) {
                    displayError = false;
                }
            }
            if (displayError) {
                errors.write(String.format("%s\n", ex.getMessage()));
            }
        }
    }

    public ArrayList<String> getComments () {
        return comments;
    }

    public void warning(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void error(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void fatalError(SAXParseException exception) throws SAXException {
        String msg = String.format("%s:%d.%d %s (in Annotations.java)", filename, locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        Main.writeError(msg);
        throw new SAXException("Boom!");
    }

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    public void startDocument() throws SAXException {
    }

    public void endDocument() throws SAXException {
    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    public void endPrefixMapping(String prefix) throws SAXException {
    }

    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        currentElement = qName;
        if (qName.equals("annotations")) {
            String page = atts.getValue("page");
            if (page == null) {
                Main.writeError(String.format("%s missing annotations.page attribute (in Annotations.java)", filename));
            }
        } else if (qName.equals("annotation")) {
            annotationNumber++;
        }
    }


    public void endElement(String uri, String localName, String qName) throws SAXException {
        if (qName.equals("annotation")) {
            String s = buffer.toString().trim();
            if (s.startsWith("f.") || s.startsWith("F.")) {
                if (!s.equals(currentFolio)) {
                    currentFolio = s;
                    anupamaComments.add(s);
                    //System.out.println(s);
                }
            }
            buffer.setLength(0);
        }
    }

    public void characters(char[] ch, int start, int length) throws SAXException {
        if (currentElement.equals("annotation")) {
            String s = new String(ch, start, length);
            if (s.contains("&")) {
                s = fixup(s);
            }
            buffer.append(s.trim());
        }
    }

    private String fixup (String s) {
        return s;                           // write this if necessary
    }
}
