package folio;

import java.io.File;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.SAXParser;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class Catalog implements SAXParserHandler
{
    static Regex re = new Regex("", "UPenn/mscoll390_item(\\d+)");
    static Hashtable<String, ManuscriptInfo> manuscripts = new Hashtable<String, ManuscriptInfo>();
    private Locator locator;
    private String filename;
    private String shortname;
    private SAXParser saxParser;
    private StringBuilder buffer = new StringBuilder();
    private boolean directoryName = false;
    private boolean textType = false;
    private boolean useManuscript = true;
    private ArrayList<SectionInfo> currentPageInfo = null;

    public Catalog () {
        saxParser = new SAXParser(this);
    }

    // This just reads the catalog.xml file and creates a ManuscriptInfo object for each manuscript element.
    // The ManuscriptInfo object is initialized later using information in the pages.xml file
    public void parse(String filename) throws Exception {
        this.filename = filename;
        buffer.setLength(0);
        shortname = new File(filename).getName();
        shortname = shortname.substring(0, shortname.indexOf("."));
        try {
            saxParser.parse(filename);
        } catch (Exception ex) {

        }
    }

    public void warning(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void error(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void fatalError(SAXParseException exception) throws SAXException {
        String msg = String.format("%s:%d.%d %s", filename, locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        Main.writeError(msg);
        throw new SAXException("Boom!");
    }

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    public void startDocument() throws SAXException {
    }

    public void endDocument() throws SAXException {
    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    public void endPrefixMapping(String prefix) throws SAXException {
    }

    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        if (qName.equals("manuscript")) {
            useManuscript = true;
        } else if (qName.equals("manuscriptX")) {
            useManuscript = false;
        } else if (qName.equals("directoryName")) {
            directoryName = true;
        } else if (qName.equals("textType")) {
            textType = true;
        }
    }


    public void endElement(String uri, String localName, String qName) throws SAXException {
        if (qName.equals("directoryName")) {
            currentPageInfo = new ArrayList<SectionInfo>();
            String _directoryName = buffer.toString().trim();
            if (useManuscript) {
                String actualName = null;
                manuscripts.put(_directoryName, new ManuscriptInfo(_directoryName, currentPageInfo));
            }
            buffer.setLength(0);
            directoryName = false;
        } else if (qName.equals("textType")) {
            String _textType = buffer.toString().trim();
            currentPageInfo.add(new SectionInfo(_textType));
            if (_textType.endsWith("Supp")) {
                SectionInfo mainText = currentPageInfo.get(currentPageInfo.size()-2);
                SectionInfo suppText = currentPageInfo.get(currentPageInfo.size()-1);
                suppText.files = mainText.files;                    // point Supp text to same pages as main text
            }
            textType = false;
            buffer.setLength(0);
        }
    }

    public void characters(char[] ch, int start, int length) throws SAXException {
        if (directoryName || textType) {
            buffer.append(new String(ch, start, length));
        }
    }
    
    @Override
    public String toString () {
        StringBuilder sb = new StringBuilder();
        Enumeration<String> en = Catalog.manuscripts.keys();
        
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            ManuscriptInfo mi = Catalog.manuscripts.get(key);
            sb.append(mi.toString());
            sb.append("\n");
        }
        return sb.toString();
    }

}
