package folio;

// This does nothing. It is a template for writing a class that parses XML

import java.io.BufferedWriter;
import java.io.File;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.SAXParserHandler;
import org.sanskritlibrary.TextFormatter;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
//DefaultHandler implements ContentHandler, DTDHandler, EntityResolver, ErrorHandler
//import org.xml.sax.helpers.DefaultHandler;

//DefaultHandler2 implements ContentHandler, DTDHandler, EntityResolver, ErrorHandler, DeclHandler, EntityResolver2, LexicalHandler
//http://docs.oracle.com/javase/1.5.0/docs/api/org/xml/sax/ext/DefaultHandler2.html
import org.xml.sax.ext.DefaultHandler2;         // this allows processing of comments because it implements LexicalHandler


// SAXParserHandler is an interface that forces us to include required methods.
// Serializable is optional. Have made local variables that don't need to be serialized transient.
public class FixupXML extends DefaultHandler2 implements SAXParserHandler, Serializable {
    static int debugCount = 0;
    transient Locator locator;
    transient ArrayList<String> pathStack = new ArrayList<String>();
    transient Hashtable<String, StringBuilder> buffers = new Hashtable<String, StringBuilder>();
    ManuscriptInfo mi;
    String filename;
    SAXParser saxParser;
    BufferedWriter bw;
    boolean endCommentWithEOL = true;
    int charactersLength = 0;
    boolean ignore = false;
    ArrayList<ParseState> XMLCounter = new ArrayList<ParseState>();
    int currentXMLCounter = 0;
    Hashtable<String, Hashtable<String, String>> attributes = new Hashtable<String, Hashtable<String, String>>();
    Hashtable<String, String> reusableHash = new Hashtable<String, String>();
    private StringBuilder buffer = new StringBuilder();
    private boolean endOfStartElementPending = false;           // for comments
    private boolean doDumpBuffer = true;
    private boolean inTextElement = true;

    public FixupXML (String filename, ManuscriptInfo mi) throws Exception
    {
        this.filename = filename;
        this.mi = mi;
        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();

        saxParser = factory.newSAXParser();
        saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
    }

    public void parse () throws Exception {
        try {
            saxParser.parse( new File(filename), this);
        } catch (Exception ex) {
            throw ex;
        }

    }

	// Maintains a stack of element names (qName)
    private void push (String qName) {
        pathStack.add(qName);
    }

	// Pop a qName from the stack
    private String pop () {
        int last = pathStack.size()-1;
        String top = null;
        if (last >= 0) {
            top = pathStack.get(last);
            pathStack.remove(last);
        }

        return top;
    }

	// Peek at a qName without popping stack
    private String peek () {
        int last = pathStack.size()-1;
        String top = null;
        if (last >= 0) {
            top = pathStack.get(last);
        }

        return top;

    }

	// Serialize the qName stack. The string that is returned represent the path to the current qName
    private String buildPath () {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (String name : pathStack) {
            sb.append(String.format("%s%s", delim, name));
            delim = ".";
        }
        return sb.toString();
    }

    private void displayError(String msg) {
        System.out.format("Line %d: %s\n", locator.getLineNumber(), msg);
    }

    @Override
    public void warning(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void error(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    @Override
    public void startDocument() throws SAXException {
        File f = new File(Main.dirXMLCatalogWithFacsimiles);
        f.mkdirs();
        String out = String.format("%s%s.xml", Main.dirXMLCatalogWithFacsimiles, mi.catalogName);
        try {
            bw = IOAssistance.openWriter(out);
            bw.write("<?xml version='1.0' encoding='UTF-8'?>\n");
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }

    @Override
    public void endDocument() throws SAXException {
        try {
            bw.write(buffer.toString());        // in case there are any comments in the buffer
            bw.close();
        // Now check for unaccounted for zones
        if (mi.insertionPoints.size() > 1) {
            for (SurfaceInfo si : mi.surfacesRepaired) {
                int numZones = si.zones.size();
                for (int i=0; i<numZones; i++) {
                    if (!si.zones.get(i).accountedFor) {
                        Main.writeError(String.format("%s.%s.zone[%d]\n", mi.catalogName, si.graphic.url, i));
                    }
                }
            }
        }

        } catch (Exception ex) {

        }
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException {
    }


    private void dumpBuffer(String qName) throws SAXException {
        if (!doDumpBuffer) {
            buffer.setLength(0);
            doDumpBuffer = true;
            return;
        }
        try {
            boolean writeEnd = false;
            ParseState ps = null;
            if (XMLCounter.size() > 0) {
                ps = XMLCounter.get(XMLCounter.size()-1);
            }
            if (qName == null) {        // called from startElement, have not encountered end yet so close previous element with a >
                if (ps != null) {        // make sure there is an open element
                    if (endOfStartElementPending && ps.elementCount == 0) {
                        bw.write(">");          // note this will be followed by whatever is in buffer
                    }
                }
            } else {    // called from endElement need to see if we need /> or </qName>
                int xmlLength = currentXMLCounter - ps.xmlCount;
                if (xmlLength == 0) {   // no content after start of element
                    bw.write("/>");     //
                } else {
                    if (endOfStartElementPending && ps.elementCount == 0) {
                        bw.write(">");
                    }
                    writeEnd = true;
                }
            }

            if (buffer.length() > 0) {
                String s = buffer.toString();
                s = s.replaceAll("& ", "&amp; ");
                bw.write(s);
                buffer.setLength(0);
            }
            if (writeEnd) {
                bw.write(String.format("</%s>", qName));
            }
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        if (ignore) {
            return;
        }

        if (qName.equals("text")) {
            inTextElement = true;
        }

        dumpBuffer(null);
        if (XMLCounter.size() > 0) {
            XMLCounter.get(XMLCounter.size()-1).elementCount++;   // keep track of intervening elements
        }

        endCommentWithEOL = false;
        currentXMLCounter++;
        ParseState ps = new ParseState();
        ps.xmlCount = currentXMLCounter;
        XMLCounter.add(ps);
        if (qName.equals("facsimile")) {
            Hashtable<String, String> h = new Hashtable<String, String>();
            copyAttributes(h, "facsimile", atts);
            ignore = true;
        } else {
            boolean doWrite = true;
            if (inTextElement) {
                if (qName.equals("note") || qName.equals("bibl") || qName.equals("desc")) {
                    if (atts.getValue("xml:lang") == null) {
                        reusableHash.clear();
                        copyAttributes(reusableHash, qName, atts);
                        reusableHash.put("xml:lang", "en");
                        writeStartElement(qName, reusableHash);
                        doWrite = false;
                    }
                }
            }
            if (doWrite) {
                writeStartElement(qName, atts);
            }
        }
    }


    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        if (!ignore) {
            writeEndElement(qName);
            XMLCounter.remove(XMLCounter.size()-1);
        }
        
        if (qName.equals("text")) {
            inTextElement = false;
        }
        
        if (qName.equals("facsimile")) {
            ignore = false;
            insertSurfaces();
        }
    }


    private void copyAttributes(Hashtable<String, String> h, String key, Attributes atts) {
        attributes.put(key, h);
        int len = atts.getLength();
        for (int i=0; i<len; i++) {
            String name = atts.getQName(i);
            String value = atts.getValue(i);
            h.put(name, value);
        }
    }


    private void writeStartElement(String qName, Attributes atts) throws SAXException {
        reusableHash.clear();
        copyAttributes(reusableHash, qName, atts);
        writeStartElement(qName, reusableHash);
    }

    private void writeStartElement (String qName, Hashtable<String, String> h) throws SAXException {
        try {
            String s = String.format("<%s", qName);
            bw.write(s);
            if (h.size() > 0) {
                bw.write(" ");
            }
            Enumeration<String> en = h.keys();
            String delim = "";
            while (en.hasMoreElements()) {
                String name = en.nextElement();
                String value = h.get(name);
                s = String.format("%s%s='%s'", delim, name, value);
                delim = " ";
                bw.write(s);
            }
            endOfStartElementPending = true;            // for comments
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }

    private void writeEndElement(String qName) throws SAXException {
        try {
            dumpBuffer(qName);
        } catch (Exception ex) {
        }
    }

    private void insertSurfaces () throws SAXException {
        try {
            if (filename.contains("2174")) {
                filename = filename + "";
            }
            Hashtable<String, String> h = attributes.get("facsimile");
            writeStartElement("facsimile", h);
            bw.write(">\n");              // need to close this ourselves
            charactersLength++;         // to force </facsimile>
            String xmlid = h.get("xml:id");
            boolean found = false;
            for (InsertionPoint ip : mi.insertionPoints) {
                debugCount++;
                if (ip.getID().equals(xmlid)) {
                    insertSurfaces(ip);
                    found = true;
                    break;
                }
            }
            if (!found) {
                Main.writeError(String.format("No locus found with target='#%s'", xmlid));
            }
            bw.write("</facsimile>\n");                // > will be provided in usual way
            doDumpBuffer = false;
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }

    private void insertSurfaces(InsertionPoint ip) throws Exception {
        // find first surface
        int slen = mi.surfacesRepaired.size();
        int i = 0;
        String facsGrp = ip.getID();
        int firstSurface = -1;
        int firstZone = -1;
        int lastSurface = -1;
        int lastZone = -1;
        String from = ip.getFrom();
        String to = ip.getTo();
        if (from == null) {
            Main.writeError(String.format("%s: locus that references facsimile %s does not have a from attribute",
                    mi.directoryName, ip.getID()));
            return;
        }

        if (mi.insertionPoints.size() == 1) {       // if only one facsimile section put all surfaces in it
            for (int j=0; j<slen; j++) {
                SurfaceInfo si = mi.surfacesRepaired.get(j);
                    bw.write(si.toString());
                    bw.write("\n");
            }
            return;
        }

        // Reach here if more than one facsimile section and we have to detect crops that fall through the cracks
        if (!from.startsWith("f. ")) {
            from = "f. " + from;
        }
        if (to != null && !to.startsWith("f. ")) {
            to = "f. " + to;
        }
        // Find first surface
        for (i = 0; i < slen && firstSurface == -1; i++) {
            SurfaceInfo si = mi.surfacesRepaired.get(i);
            int zlen = si.zones.size();
            for (int j = 0; j < zlen && firstSurface == -1; j++) {
                ZoneInfo zi = si.zones.get(j);
                if (zi.graphic.desc == null) {
                    continue;       // caused by an error reported elsewhere. File is not in pages.xml but was sent to Donglei anyway
                } else {
                    if (zi.graphic.descNormalized == null) {
                        zi.graphic.descNormalized = TextFormatter.normalizeID(zi.graphic.desc);
                    }
                    if (zi.graphic.descNormalized.equals(from)) {
                        firstSurface = i;
                        firstZone = j;
                    }
                }
            }
        }
        
        if (slen > 0) {
            i--;        // in case the zone of the last surface is in same surface as the zone of the first surface
        }
        
        // Find last surface
        for (; i < slen && lastSurface == -1; i++) {
            SurfaceInfo si = mi.surfacesRepaired.get(i);
            int zlen = si.zones.size();
            for (int j = 0; j < zlen && lastSurface == -1; j++) {
                ZoneInfo zi = si.zones.get(j);
                if (zi.graphic.desc == null) {
                    continue;           // caused by an error reported elsewhere. File is not in pages.xml but was sent to Donglei anyway
                } else {
                    if (zi.graphic.descNormalized == null) {
                        zi.graphic.descNormalized = TextFormatter.normalizeID(zi.graphic.desc);
                    }
                    if (to != null && zi.graphic.descNormalized.equals(to)) {
                        lastSurface = i;
                        lastZone = j;
                        break;
                    }
                }
            }
        }
        if (lastSurface == -1) {
            String msg = String.format("\n%s: %s could not find surfaces for %s--%s\n\tformula=%s",
                    mi.directoryName,
                    ip.getID(),
                    ip.getFrom() == null ? "???" : ip.getFrom(),
                    ip.getTo() == null ? "???" : ip.getTo(),
                    mi.foliationFormula);
            Main.writeError(msg);
        } else {
            for (int j=0; j<slen; j++) {
                SurfaceInfo si = mi.surfacesRepaired.get(j);
                int numZones = si.zones.size();
                if ((si.facsGrp != null && si.facsGrp.equals(facsGrp)) || (j>=firstSurface && j<=lastSurface)) {
                    boolean processed = false;
                    if (j == firstSurface) {
                        for (int k=firstZone; k<numZones; k++) {
                            si.zones.get(k).accountedFor = true;
                        }
                        processed = true;
                    }
                    if (j == lastSurface) {
                        for (int k=0; k<=lastZone; k++) {
                            si.zones.get(k).accountedFor = true;
                        }
                        processed = true;
                    }
                    if (!processed) {
                        for (int k=0; k<numZones; k++) {
                            si.zones.get(k).accountedFor = true;
                        }
                    }

                    bw.write(si.toString());
                    bw.write("\n");
                }
            }
        }

        
    }

    @Override
    // Collects the innerText in a StringBuilder keyed by the current path. It is endElement's responsibility
    // to retrieve the StringBuilder and set its length to 0.
    // This is optional. It is not specified in SAXParserHandler
    public void characters(char[] ch, int start, int length) throws SAXException {
        if (!ignore) {
            currentXMLCounter++;            // register the presence of some content
            String s = new String(ch, start, length);

            buffer.append(s);
            charactersLength += s.length();
        }
    }

    @Override
    public void processingInstruction (String name, String args) throws SAXException {
        try {
            String s = String.format("<?%s %s?>\n", name, args);
            bw.write(s);
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }
    @Override
    // Optional if you want to be notified of comments
    public void comment(char[] ch, int start, int length) throws SAXException {
        try {
            String s = String.format("<!--%s-->", new String(ch, start, length));
            buffer.append(s);
            if (endCommentWithEOL) {
               buffer.append("\n");
            }
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }

    class ParseState {
        int xmlCount;
        int elementCount;
    }

}
