package folio;

import java.io.File;
import java.util.ArrayList;
import org.sanskritlibrary.SAXParser;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

// Note: this parses pages.xml which should be named images.xml
public class Images implements SAXParserHandler
{
    private ArrayList<String> comments = new ArrayList<String>();
    private Locator locator;
    private String filename;
    private String shortname;
    private SAXParser saxParser;
    private StringBuilder buffer = new StringBuilder();
    private ArrayList<SectionInfo> currentImageInfo;
    private String currentTextType;
    private int normalPageCount = 0;
    private String currentManuscriptName;
    private int imageCount = 0;

    public Images () {
        saxParser = new SAXParser(this);
    }

    public void parse(String filename) throws Exception {
        this.filename = filename;
        buffer.setLength(0);
        shortname = new File(filename).getName();
        shortname = shortname.substring(0, shortname.indexOf("."));
        try {
            saxParser.parse(filename);
        } catch (Exception ex) {
            System.out.format("pages.xml(%s) failed parse with error %s\n", filename, ex.getMessage());
        }
    }

    public ArrayList<String> getComments () {
        return comments;
    }

    public void warning(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void error(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void fatalError(SAXParseException exception) throws SAXException {
        String msg = String.format("%s:%d.%d %s (in Images.java)\n", filename, locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        Main.writeError(msg);
        throw new SAXException("Boom!");
    }

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    public void startDocument() throws SAXException {
    }

    public void endDocument() throws SAXException {
    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    public void endPrefixMapping(String prefix) throws SAXException {
    }

    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        if (qName.equals("directory")) {
            imageCount = 0;
            String _normalPageCount = atts.getValue("normalPageCount");
            if (_normalPageCount == null) {
                _normalPageCount = "1";
            }
            normalPageCount = Integer.parseInt(_normalPageCount);
            currentManuscriptName = atts.getValue("name");

            if (Catalog.manuscripts.get(currentManuscriptName) != null) {
                currentImageInfo = Catalog.manuscripts.get(currentManuscriptName).imageInfo;
                currentTextType = currentImageInfo.get(0).textType;          // this is the default. It can be overridden by the textType
            } else {
                currentImageInfo = null;
                Main.writeError(String.format("%s not in catalogs.xml (in Images.java)", currentManuscriptName));
            }
        } else if (qName.equals("page")) {                  // page is not a good name, image would be better.
            if (currentImageInfo == null) {             // this error has already been reported above
                return;
            }
            imageCount++;
            boolean newTextType = false;
            if (atts.getValue("textType") != null) {
                currentTextType = atts.getValue("textType");
                newTextType = true;
            }
            int numPages = 0;
            boolean anomaly = false;

            String name = null;
            if (atts.getValue("anomaly") != null) {
                anomaly = true;
                numPages = Integer.parseInt(atts.getValue("anomaly"));
            }
            else {
                numPages = normalPageCount;
            }

            name = atts.getValue("name");
            boolean found = false;
            for (SectionInfo pi : currentImageInfo) {
                if (pi.textType.equals(currentTextType)) {
                    FileInfo fi = new FileInfo(name);
                    fi.numFolioPages = numPages;
                    fi.facsid = atts.getValue("facsid");
                    fi.facsGrp = atts.getValue("facsGrp");
                    pi.files.add(fi);
                    found = true;
                    break;
                }
            }
            if (!found && newTextType) {
                Main.writeError(String.format("pages.xml:%d, textType %s not found (in Images.java)",
                            locator.getLineNumber(), currentTextType));
            }
        }
    }



    public void endElement(String uri, String localName, String qName) throws SAXException {
    }


    public void characters(char[] ch, int start, int length) throws SAXException {
    }

    private String fixup (String s) {
        return s;                           // write this if necessary
    }
}
