/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package folio;

import java.io.BufferedWriter;
import java.io.File;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.Config;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;

/**
 *
 * @author ralph
 */
public class Main {
    static int debug = 0;
    static private Regex rePageName = new Regex("", "(\\d+[ab]?.jpg)$");
    static private Regex reAnnotations = new Regex("", "(annotations.*)");
    static private Regex reCropFile = new Regex("", "(.*)crops_tei.xml$");
    static String dirCatalog;               // full name of catalog.xml
    static String dirImagesLayout;          // pages.xml
    static String dirAnnotations;           // full name of the annotations directory
    static String dirXMLCatalog;                 // full name of the directory containing Peter's xml files
    static String dirXMLCatalogWithFacsimiles;   // full name of the directory containing Peter's xml files with facsimiles added
    static String dirXMLZones;                 // full name of the directory containing Donglei's xml files
    static String dirXMLZonesRepaired;
    static String dirCropped;               // full name of the directory containing the cropped images
    static String dirOriginal;                  // full name of the directory containing the whole images
    static String htmlDirectory = "folioHTML/";
    static BufferedWriter pf;   // progress file
    static BufferedWriter errorFile;
    static BufferedWriter infoFile;
    static final Hashtable<String, Boolean> ignore = new Hashtable<String, Boolean>();
    static {
        ignore.put("IndicSubjectClassification.xml", Boolean.TRUE);
        ignore.put("listbiblfile.xml", Boolean.TRUE);
        ignore.put("listpersonfile.xml", Boolean.TRUE);
        ignore.put("neh2008.xml", Boolean.TRUE);
    }
    static boolean usePetite = false;           // use original Donglei image

   
    @SuppressWarnings("ResultOfObjectAllocationIgnored")
    public static void main(String[] args) throws Exception {
        errorFile = IOAssistance.openWriter("folioErrorReport.txt");
        infoFile = IOAssistance.openWriter("infoReport.txt");
        Annotations.errors = IOAssistance.openWriter("anupamaErrors.txt");
        pf = IOAssistance.openWriter("progress.txt");

        Main.progress("The following three files should be inspected when the program finishes");
        Main.progress("  1. infoReport.txt. Displays the foliation formula for each manuscript and how the folios are distributed across the facsimile sections");
        Main.progress("  2. anupamaErrors.txt. Will contains errors in Anupama's annotations that were not forgiven in knownErrors.txt\n");
        Main.progress("  3. folioErrorReport.txt. Will contain errors that occurred while trying to match a foliation formula with cropped images.");

        /*
        System.out.println("This program creates the following files:\n");
        System.out.println("1. errorReport.txt\n");
        System.out.println("2. anupamaErrors.txt\n");
        System.out.println("3. repair.sh (reverses flipped crops. Must be given execute permission using chmod ug+x\n");
         */
        new Main(args);
        errorFile.close();
        infoFile.close();
        Annotations.errors.close();
        pf.close();
    }



    public static void writeError (String msg) {
        try {
            errorFile.write(msg);
            if (!msg.endsWith("\n")) {
                errorFile.write("\n");
            }
        } catch (Exception ex) {
            System.out.println(msg);
        }
    }

    public static void writeInfo (String msg) {
        try {
            infoFile.write(msg);
            if (!msg.endsWith("\n")) {
                infoFile.write("\n");
            }
        } catch (Exception ex) {
            System.out.println(msg);
        }
    }

    public static void progress (String msg) {
        try {
            pf.write(msg);
            pf.write("\n");
        } catch (Exception ex) {
            System.out.format("Error writing progress.txt");
        }
    }
    
    private Main (String[] args) throws Exception {
        boolean rc = true;
        boolean croppingsOnly = true;
        int numargs = args.length;
        usePetite = false;
        for (int i=0; i<numargs; i++) {
            String arg = args[i];
            if (arg.equals("includeOriginal")) {
                croppingsOnly = false;
            } else if (arg.equals("usePetite")) {
                usePetite = true;
            } else {
                System.out.format("%s is not a known parameter, ignored\n", arg);
            }
        }

        rc = readPaths();
        if (!rc) {
            return;
        }
        
        Errors errors = new Errors();
        errors.parse("knownErrors.xml");
        
        Main.progress("knownErrors.xml read to discover what errors to ignore\n");

        Main.progress(String.format("Opening %s\n", dirCatalog));
        Catalog catalog = new Catalog();            
        catalog.parse(dirCatalog);             

        Images pages = new Images();
        pages.parse(dirImagesLayout);       // this updates the static Catalog.manuscripts hash table

        collectAnupamaComments();
        getFoliationFormulas();

        //dumpManuscripts();

        verifyFoliationFormulas();

        BufferedWriter repairScript = IOAssistance.openWriter("repair.sh");
        System.out.format("Getting zone xml files from %s\n", dirXMLZones);
        getZones(repairScript, new File(dirXMLZones), "");
        repairScript.close();

        addSurfaces();

        /*
         At this point each ManuscriptInfo object has the following information
         * 1. A List containing Anupama's comments on each image in the manuscript
         * 2. A List of FileInfo for each image file of the manuscript
         * 3. A List of SectionInfo describing the files comprising each text type in the manuscript
         * 4. A List of insertion points for inserting surfaces into catalog files.
         * 5. A List of folio IDs for the pages of the images of the manuscript
         * 6. A List of surfaces (cropped images) for the images of the manuscript
         * This information is presented in HTML pages next.
         */
        createHTMLIndex();
        createHTMLPages(croppingsOnly);        
    }

    private void getFoliationFormulas () throws Exception {
        XMLCatalog xmlCatalog = new XMLCatalog();
        File[] files = new File(dirXMLCatalog).listFiles();
        for (File file : files) {
            String filename = file.getName();
            if (!filename.startsWith(".") && filename.endsWith(".xml") && ignore.get(filename) == null) {
                xmlCatalog.parse(filename);

                String key = filename.substring(0, filename.length() - 4);
                String dirkey = MapName.mapCatalogName.get(key);
                if (dirkey != null) {
                    ManuscriptInfo mi = Catalog.manuscripts.get(dirkey);
                    if (mi == null) {
                        writeError(String.format("In Main.getFoliationFormulas: %s is mentioned in catalog.xml but was not found. (in Main.java)", key));
                    } else {
                        mi.setFoliationFormula(xmlCatalog.getFormula());
                        mi.setInsertionPoints(xmlCatalog.getInsertionPoints());
                    }
                }
            }
        }
    }

    private void getZones(BufferedWriter repairScript, File dir, String dirname) throws Exception {
        XMLZones xmlZones = new XMLZones(repairScript);
        String tempfile = "tempzone.xml";
        File[] files = dir.listFiles();
        for (File file : files) {
            if (file.isDirectory() && !file.getName().equals("Repaired")) {     // avoid Repaired directory because its files have been repaired.
                getZones(repairScript, file, file.getName());
            } else {
                String filename = file.getName();
                if (!filename.startsWith(".") && filename.endsWith(".xml") && ignore.get(filename) == null) {
                    // create root element
                    System.out.format("Reading %s\n", file.getAbsoluteFile());
                    String xml = IOAssistance.readEntireFile(file.getAbsolutePath());
                    String[] lines = xml.split("\n");
                    BufferedWriter bw = IOAssistance.openWriter(tempfile);
                    int count = 0;
                    for (String line : lines) {
                        if (count == 0) {
                            bw.write("<?xml version='1.0' encoding='UTF-8' ?>\n");
                            bw.write("<zones>\n");
                            if (!line.startsWith("<?xml")) {
                                bw.write(line);
                            }
                        } else {
                            bw.write(line);
                            bw.write("\n");
                        }
                        count++;
                    }
                    bw.write("</zones>\n");
                    bw.close();

                    String dirkey = null;
                    reCropFile.reset(filename);
                    if (reCropFile.find()) {
                        dirkey = String.format("%s/%s", dirname, reCropFile.getGroup(1));
                    } else if (!filename.contains("notInCatalog")){
                        writeError(String.format("Crop file %s cannot be matched with a catalog file", filename));
                        continue;
                    }

                    if (dirkey != null) {
                        String imageDirectory = null;
                        if (usePetite) {
                            imageDirectory = String.format("%s%s/split/", dirCropped, dirkey).substring(7);
                        } else {
                            imageDirectory = String.format("%s%s/", dirCropped, dirkey).substring(7);
                        }
                        ManuscriptInfo mi = Catalog.manuscripts.get(dirkey);
                        if (mi == null) {
                            writeError(String.format("Catalog %s not found for zones file %s", dirkey, filename));
                        } else {
                            mi.getPageIDs();
                            xmlZones.parse(dir.getName(), tempfile, file.getAbsolutePath(), imageDirectory, mi);
                        }
                    }
                }
            }
        }
    }

    private void dumpManuscripts () {
        Enumeration<String> en = Catalog.manuscripts.keys();

        while (en.hasMoreElements()) {
            String key = en.nextElement();
            ManuscriptInfo mi = Catalog.manuscripts.get(key);
            writeError(String.format("%s: %s (in Main.java)", key, mi.toString()));
        }
    }



    private void collectAnupamaComments() throws Exception {
        boolean firstTime = true;
        Enumeration<String> en = Catalog.manuscripts.keys();
        Annotations annotations = new Annotations();
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            ManuscriptInfo mi = Catalog.manuscripts.get(key);
            for (SectionInfo pi : mi.imageInfo) {
                String textType = pi.textType;
                if (!textType.endsWith("Supp")) {
                    String annodir = String.format("%s%s/%s/", dirAnnotations, textType, mi.directoryName);
                    if (firstTime) {
                        Main.progress(String.format("Attempting to find annotations for each manuscript defined in catalog.xml"));
                        Main.progress(String.format("The path to an annotation directory is %s<textType>/<directoryName>/", dirAnnotations));
                    }
                    reAnnotations.reset(annodir);

                    if (reAnnotations.find()) {
                        String g = reAnnotations.getGroup(1);
                        if (Errors.missingAnnotationDirectories.get(g) != null ||
                            Errors.emptyAnnotationDirectories.get(g) != null) {
                            if (firstTime) {
                                firstTime = false;
                                Main.progress("Note: at least one annotation directory was skipped because it was mentioned in knownErrors.xml");
                            }
                            continue;
                        }
                    }
                    if (new File(annodir).exists() == false) {          // should never happen
                        reAnnotations.reset(annodir);
                        if (reAnnotations.find()) {
                            writeError(String.format("%s does not exist (in Main.java)", annodir));
                        }
                    } else {
                        for (FileInfo fi : pi.files) {
                            if (fi.name.equals("mscoll390_item491_wk1_front0001.jpg")) {
                                debug++;
                            }
                            debug++;
                            if (debug == 2240) {
                                debug = 2240;
                            }

                            String filename = fi.name + ".xml";
                            if (new File(annodir + filename).exists() == false) {
                                boolean matchFound = false;
                                for (Regex re : Errors.patterns) {
                                    re.reset(fi.name);
                                    if (re.find()) {
                                        matchFound = true;
                                        break;
                                    }
                                }

                                if (matchFound) {
                                    filename = fi.name + ".xml";
                                } else {
                                    rePageName.reset(fi.name);
                                    if (rePageName.find()) {
                                        filename = rePageName.getGroup(1);
                                        filename = filename.replace(".jpg", ".xml");
                                    } else {
                                        filename = fi.name + ".xml";
                                    }
                                }
                            }
                            annotations.parse(String.format("%s%s", annodir, filename), fi.anupamaComments);
                        }
                    }
                }
            }
            // Reach here when we have all of Anupama's comments for manuscript mi.
            //Fixup up situations where the same image has parts of two different texts.
            mi.consolidateComments();
        }
    }


    private void verifyFoliationFormulas () throws Exception {
        Enumeration<String> en = Catalog.manuscripts.keys();

        writeError("\n\nEach folio contains recto and verso pages.The file pages.xml describes\nhow many pages there are on each image using the normalPageCount and anomaly attributes.\nThe number of pages computed from page.xml should equal twice the number of folios.\nThe manuscripts which violate this constraint are listed next.\n\n");

        while (en.hasMoreElements()) {
            String key = en.nextElement();
            ManuscriptInfo mi = Catalog.manuscripts.get(key);
            if (!mi.verifyFoliation()) {
                
            }
        }
    }

    private void addSurfaces () throws Exception {
        writeError(String.format("\n\nSurfaces are automatically added to catalog files in the directory %s\n", dirXMLCatalog));
        writeError(String.format("The updated files are written to the directory %s/WithFacsimiles\n", dirXMLCatalog));
        writeError(String.format("Errors that occurred are listed below. These errors are of two types\n"));
        writeError(String.format("1. The number of pages is not double the number of folios. This error was reported earlier\n"));
        writeError(String.format("2. Either the from or to attribute is incorrect.\n"));

        Enumeration<String> en = ignore.keys();
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            String filename = String.format("%s%s", dirXMLCatalog, key);
            String text = IOAssistance.readEntireFile(filename);
            new File(dirXMLCatalogWithFacsimiles).mkdirs();
            filename = String.format("%s%s", dirXMLCatalogWithFacsimiles, key);
            BufferedWriter bw = IOAssistance.openWriter(filename);
            bw.write(text);
            bw.close();
        }
        ManuscriptInfo.errorCache.clear();
        en = Catalog.manuscripts.keys();

        while (en.hasMoreElements()) {
            String key = en.nextElement();
            if (key.contains("2639")) {
                //System.out.println("x");
            }
            ManuscriptInfo mi = Catalog.manuscripts.get(key);
            mi.addSurfaces();
        }

        writeError(String.format("\n\nThe following .xml files were not found. If they were not done please comment out the manuscript in catalog.xml"));
        for (String error : ManuscriptInfo.errorCache) {
            writeError(String.format("\t%s", error));
        }
    }

    private void createHTMLIndex () throws Exception {
        StringBuilder sb = new StringBuilder();
        sb.append("<html>\n<head>\n</head>\n<body>\n");
        BufferedWriter bw = null;
        for (String cn : MapName.sorted) {
            cn = MapName.mapCatalogName.get(cn);
            ManuscriptInfo mi = Catalog.manuscripts.get(cn);
            if (mi == null) {
                continue;
            }
            String row = mi.generateIndexPageEntry();
            if (row != null) {
                sb.append(row);
            }
        }
        sb.append("</body>\n</html>\n");
        File f = new File(htmlDirectory);
        f.mkdirs();
        bw = IOAssistance.openWriter(String.format("%sindex.html", htmlDirectory));
        bw.write(sb.toString());
        bw.close();
    }

    private void createHTMLPages (boolean croppingsOnly) throws Exception {
        Enumeration<String> en = Catalog.manuscripts.keys();
        File f = new File(htmlDirectory);
        if (!f.exists()) {
            f.mkdir();
        }

        while (en.hasMoreElements()) {
            String key = en.nextElement();
            ManuscriptInfo mi = Catalog.manuscripts.get(key);
            String html = mi.generateHTML(croppingsOnly, usePetite);
            if (html != null) {
                BufferedWriter bw = IOAssistance.openWriter(String.format("%s%s.html", htmlDirectory, mi.catalogName));
                bw.write(html);
                bw.close();
            }
        }
    }


    private boolean fileExists (String name) {
        File f = new File(name);
        if (f.exists()) {
            return true;
        } else {
            System.out.format("%s does not exist.\n", name);
            return false;
        }
    }

    
    private boolean readPaths () throws Exception {
        boolean pathsRead = true;
        Config config = new Config("config.xml");
        //dumpConfig(config);
        dirCatalog = config.masterCatalog;
        pathsRead &= fileExists(dirCatalog);

        dirImagesLayout = config.imagesLayout;
        pathsRead &= fileExists(dirImagesLayout);

        dirAnnotations = config.imageAnnotations;
        pathsRead &= fileExists(dirAnnotations);

        dirXMLCatalog = config.catalogEntriesOriginal;
        pathsRead &= fileExists(dirXMLCatalog);

        dirXMLCatalogWithFacsimiles = config.catalogEntriesWithFacsimiles;
        //pathsRead &= fileExists(dirXMLCatalogWithFacsimiles);

        dirXMLZones = config.croppedXMLOriginal;
        pathsRead &= fileExists(dirXMLZones);

        dirXMLZonesRepaired = config.croppedXMLRepaired;
        //pathsRead &= fileExists(dirXMLZonesRepaired);

        if (usePetite) {
            dirCropped = config.imagesCroppedReduced;
            pathsRead &= fileExists(dirCropped);
            dirCropped = "file://" + dirCropped;
        } else {
            dirCropped = config.imagesCropped;
            pathsRead &= fileExists(dirCropped);
            dirCropped = "file://" + dirCropped;
        }

        dirOriginal = config.imagesSitaAnnotatedReduced;
        pathsRead &= fileExists(dirOriginal);
        dirOriginal = "file://" + dirOriginal;

        return pathsRead;
    }
}
