package folio;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.TextFormatter;

public class ManuscriptInfo {
    public static String notImaged = "notImaged.jpg";
    private static Regex reWork = new Regex ("", "^(?:([IVX]+[a-z]?):)?\\s*(.*)");
    private static Regex reNumber = new Regex("", "^(\\d+)([abc.]?)$");
    private static Regex reRoman = new Regex("", "^([ivx]+)([abc.]?)$");
    private static Regex reRemoveJPG = new Regex("", "\\.jpg$");
    private static String _reIllegalRange = "([^-])-([^-])";
    private static Regex reIllegalRange = new Regex("", _reIllegalRange);
    private static String htmlHeader;
    private static String htmlRowHeader;
    private static String htmlRowAnupama;
    private static String htmlFooter;
    public static ArrayList<String> errorCache = new ArrayList<String>();

    static String scriptTemplate;

    static {
        htmlHeader = "<html>\n<head><script type='text/javascript'>%s</script>\n</head>\n<body onload='init()' id='bd'>\n<table id='list' border='1' style='border-collapse:collapse'>\n";
        htmlRowHeader = "<tr><td id='td%d' onclick='showImage(this, %d)'><i>%s</i><br/>";
        htmlRowAnupama = "<span style='color:green'>%s</span></td></tr>\n";
        htmlFooter = "</table><canvas id='page' style='position:absolute'/></body></html>";
        try {
            scriptTemplate = IOAssistance.readEntireFile("scriptTemplate.js");
        } catch (Exception ex) {

        }
    }
    public String directoryName;
    public String catalogName;
    public String formulaReport;

    public Hashtable<String, ArrayList<String>> comments = new Hashtable<String, ArrayList<String>>();  // Comments for each image in the manuscript
    public ArrayList<FileInfo> imageFiles = new ArrayList<FileInfo>();  // one entry for each image of manuscript
    // comments.size() == imageFiles.size()

    public ArrayList<SectionInfo> imageInfo;    // imageFiles is created from this data structure
    public ArrayList<FolioItem> folioItems = null;  // built from the foliation formula for this manuscript
    public ArrayList<InsertionPoint> insertionPoints = new ArrayList<InsertionPoint>();
    public String foliationFormula;
    private int numFolios = 0;
    private int numImages = 0;
    private int numComments;
    ArrayList<String> folioIDs = null;
    public ArrayList<SurfaceInfo> surfaces = new ArrayList<SurfaceInfo>();
    public ArrayList<SurfaceInfo> surfacesRepaired = new ArrayList<SurfaceInfo>();



    public ManuscriptInfo (String directoryName, ArrayList<SectionInfo> pageInfo) {
        this.directoryName = directoryName;
        this.catalogName = MapName.mapDirectoryName.get(directoryName);
        this.imageInfo = pageInfo;
    }


    public void setInsertionPoints (Hashtable<String, InsertionPoint> insertionPointsHash) {
        Enumeration<String> en = insertionPointsHash.keys();
        while (en.hasMoreElements()) {
            insertionPoints.add(insertionPointsHash.get(en.nextElement()));
        }

        Collections.sort(insertionPoints, new InsertionPointCompare());

        Main.writeInfo(String.format("\n\n%s", directoryName));
        Main.writeInfo(String.format("\t%s", foliationFormula));

        for (InsertionPoint ip : insertionPoints) {
            Main.writeInfo(String.format("\t%s", ip.toString()));
        }
    }


    public void addSurfaces () throws Exception {
        String in = String.format("%s%s.xml", Main.dirXMLCatalog, catalogName);
        File f = new File(in);
        if (!f.exists()) {
            errorCache.add(String.format("%s: not found even though %s is in catalog.xml", in, catalogName));
            return;
        }
        FixupXML fixup = new FixupXML(in, this);
        fixup.parse();
        /*
        String withFacsimiles = String.format("%sWithFacsimiles/", Main.dirXMLCatalog);
        f = new File(withFacsimiles);
        f.mkdirs();
        String out = String.format("%s%s.xml", withFacsimiles, catalogName);
        BufferedReader br = IOAssistance.openReader(in);
        BufferedWriter bw = IOAssistance.openWriter(out);
        String line = null;
        int lineno = 0;
        int k = 0;
        InsertionPoint cip = null;
        if (insertionPoints.size() > 1) {
            k = 0;
        }
        if (insertionPoints.size() > 0) {
            cip = insertionPoints.get(k);
        }
        while ((line=br.readLine()) != null) {
            lineno++;
            if (cip != null) {
                if (lineno == cip.getStartLine()) {
                    bw.write(line);
                    bw.write("\n");
                    while ((line = br.readLine()) != null) {        // skip any contents of the facsimile element
                        lineno++;
                        if (lineno == cip.getEndLine()) {
                            break;
                        }
                    }
                    // Reach here and line equals '</facsimile>' write it after surface elements are written
                    insertSurfaces(bw, cip, insertionPoints.size());
                    k++;
                    boolean b = k < insertionPoints.size();
                    if (b) {
                        cip = insertionPoints.get(k);
                    } else {
                        cip = null;
                    }
                }
            }

            bw.write(line);
            bw.write("\n");
        }

        bw.close();

        // Now check for unaccounted for zones
        if (insertionPoints.size() > 1) {
            for (SurfaceInfo si : surfacesRepaired) {
                int numZones = si.zones.size();
                for (int i=0; i<numZones; i++) {
                    if (!si.zones.get(i).accountedFor) {
                        System.out.format("%s.%s.zone[%d]\n", catalogName, si.graphic.url, i);
                    }
                }
            }
        }
        */
    }


    // Insert Donglei's surfaces into Peter's catalog files
    private void insertSurfaces(BufferedWriter bw, InsertionPoint ip, int numFacsimiles) throws Exception {
        // find first surface
        if (catalogName.equals("penn2259")) {
            catalogName = "penn2259";
        }
        int slen = surfacesRepaired.size();
        int i = 0;
        String facsGrp = ip.getID();
        int firstSurface = -1;
        int firstZone = -1;
        int lastSurface = -1;
        int lastZone = -1;
        String from = ip.getFrom();
        String to = ip.getTo();
        if (from == null) {
            Main.writeError(String.format("%s: locus that references facsimile %s does not have a from attribute",
                    directoryName, ip.getID()));
            return;
        }

        if (numFacsimiles == 1) {       // if only one facsimile section put all surfaces in it
            for (int j=0; j<slen; j++) {
                SurfaceInfo si = surfacesRepaired.get(j);
                    bw.write(si.toString());
                    bw.write("\n");
            }
            return;
        }

        // Reach here if more than one facsimile section and we have to detect crops that fall through the cracks
        if (!from.startsWith("f. ")) {
            from = "f. " + from;
        }
        if (to != null && !to.startsWith("f. ")) {
            to = "f. " + to;
        }
        // Find first surface
        for (i = 0; i < slen && firstSurface == -1; i++) {
            SurfaceInfo si = surfacesRepaired.get(i);
            int zlen = si.zones.size();
            for (int j = 0; j < zlen && firstSurface == -1; j++) {
                ZoneInfo zi = si.zones.get(j);
                if (zi.graphic.desc == null) {
                    continue;       // caused by an error reported elsewhere. File is not in pages.xml but was sent to Donglei anyway
                } else {
                    if (zi.graphic.descNormalized == null) {
                        zi.graphic.descNormalized = TextFormatter.normalizeID(zi.graphic.desc);
                    }
                    if (zi.graphic.descNormalized.equals(from)) {
                        firstSurface = i;
                        firstZone = j;
                    }
                }
            }
        }
        // Find last surface
        for (; i < slen && lastSurface == -1; i++) {
            SurfaceInfo si = surfacesRepaired.get(i);
            int zlen = si.zones.size();
            for (int j = 0; j < zlen && lastSurface == -1; j++) {
                ZoneInfo zi = si.zones.get(j);
                if (zi.graphic.desc == null) {
                    continue;           // caused by an error reported elsewhere. File is not in pages.xml but was sent to Donglei anyway
                } else {
                    if (zi.graphic.descNormalized == null) {
                        zi.graphic.descNormalized = TextFormatter.normalizeID(zi.graphic.desc);
                    }
                    if (to != null && zi.graphic.descNormalized.equals(to)) {
                        lastSurface = i;
                        lastZone = j;
                        break;
                    }
                }
            }
        }
        if (lastSurface == -1) {
            String msg = String.format("\n%s: %s could not find surfaces for %s--%s\n\tformula=%s",
                    directoryName,
                    ip.getID(),
                    ip.getFrom() == null ? "???" : ip.getFrom(),
                    ip.getTo() == null ? "???" : ip.getTo(),
                    foliationFormula);
            Main.writeError(msg);
        } else {
            for (int j=0; j<slen; j++) {
                SurfaceInfo si = surfacesRepaired.get(j);
                int numZones = si.zones.size();
                if ((si.facsGrp != null && si.facsGrp.equals(facsGrp)) || (j>=firstSurface && j<=lastSurface)) {
                    boolean processed = false;
                    if (j == firstSurface) {
                        for (int k=firstZone; k<numZones; k++) {
                            si.zones.get(k).accountedFor = true;
                        }
                        processed = true;
                    }
                    if (j == lastSurface) {
                        for (int k=0; k<=lastZone; k++) {
                            si.zones.get(k).accountedFor = true;
                        }
                        processed = true;
                    }
                    if (!processed) {
                        for (int k=0; k<numZones; k++) {
                            si.zones.get(k).accountedFor = true;
                        }
                    }

                    bw.write(si.toString());
                    bw.write("\n");
                }
            }
        }

    }

    public void setFoliationFormula (String foliationFormula) {
        this.foliationFormula = foliationFormula;

        if (foliationFormula == null) {
            return;
        }

        if (folioItems == null) {
            folioItems = new ArrayList<FolioItem>();
        } else {
            folioItems.clear();
        }

        numFolios = getFolioItems(foliationFormula, folioItems);

        computeNumberOfImages();
    }

    private int getFolioItems(String foliationFormula, ArrayList<FolioItem> fis) {
        boolean dump = false;
        int folioCount = 0;
        String _foliationFormula = foliationFormula.replaceAll(";", ",");
        reIllegalRange.reset(_foliationFormula);
        if (reIllegalRange.find()) {
            Main.writeError(String.format("[XML:PMS]%s: replacing - with -- in %s (in ManuscriptInfo.java)", directoryName, foliationFormula));
            _foliationFormula = _foliationFormula.replaceAll(_reIllegalRange, "$1--$2");
        }
        String[] fields = _foliationFormula.split("\\s*,\\s*");
        String enumerationID = null;
        if (fields.length == 1 && fields[0].length() == 0) {
            Main.writeError(String.format("[XML:PMS]%s has an empty foliation formula (in ManuscriptInfo.java)",
                         directoryName));
            return 0;
        }
        for (String field : fields) {
            boolean folioError = false;
            FolioItem fi = new FolioItem();
            fis.add(fi);

            reWork.reset(field);
            if (reWork.find()) {
                if (reWork.getGroup(1) != null) {
                    enumerationID = reWork.getGroup(1).trim();
                }
                field = reWork.getGroup(2).trim();
             }
             fi.enumerationID = enumerationID;
             if (field.length() == 0) {
                 continue;
             }
             if (field.charAt(0) == '[') {
                fi.inBrackets = true;
             }

            if (!field.contains("--")) {
                fi.start = new BoundaryInfo();
                getNumber(fi.start, field);
                folioCount++;
                fi.totalFolios = 1;
            } else {
                String[] boundaries = field.split("\\s*--\\s*");
                fi.start = new BoundaryInfo();
                getNumber(fi.start, boundaries[0]);
                fi.end = new BoundaryInfo();
                getNumber(fi.end, boundaries[1]);
                if (fi.start.error || fi.end.error) {
                    fi.error = true;
                    folioError = true;
                }
                if (!folioError) {
                    fi.totalFolios = fi.end.value - fi.start.value + 1;
                    if (fi.totalFolios < 0) {
                        Main.writeError(String.format("%s: why is total folios negative?", directoryName, fi.toString()));
                    }
                    folioCount += fi.totalFolios;
                }
            }
        }

        if (dump) {
            for (FolioItem fi : fis) {
                System.out.println(fi.toString());
            }
        }

        return folioCount;
    }

    // Once imageFiles created, its size will be the number of images.
    private void computeNumberOfImages () {
        int numBlankImages = 0;
        Hashtable<String, Boolean> h = new Hashtable<String, Boolean>();
        for (SectionInfo pi : imageInfo) {
            if (!pi.textType.endsWith("Supp")) {
                for (FileInfo fi : pi.files) {
                    if (fi.name.equals(notImaged)) {            // all blank images have the same name so hash table can't be used to count them.
                        numBlankImages++;
                    } else {
                        h.put(fi.name, Boolean.TRUE);
                    }
                }
            }
        }
        numImages = h.size() + numBlankImages;
    }

    public boolean verifyFoliation() throws Exception {
        if (foliationFormula == null) {
            return true;
        }

        for (FolioItem fi : folioItems) {
            if (fi.error) {
                return false;
            }
        }


        boolean insync = imageFolioCountConsistent();

        formulaReport = String.format("(%s) #f=%d, #p=%d (should be %d)",
                foliationFormula, numFolios, computeNumberOfFolioPages(), 2*numFolios);
        if (!insync) {
            String asterisk = "";
            if (numFolios > numImages) {
                asterisk = "***";
            }
            Main.writeError(String.format("\n%s[XML:PMS/Page.xml]%s: number of pages not double the number of folios - \n\t%s (in ManuscriptInfo.java)",
                    asterisk, directoryName, formulaReport));
        }

        return true;
    }


    private int computeNumberOfFolioPages () {
        int numPages = 0;
        for (FileInfo fi : imageFiles) {
            numPages += fi.numFolioPages;            // if fi.numPages == 0 then it is not a folio page
        }
        return numPages;
    }

    private boolean imageFolioCountConsistent () {
        boolean insync = false;
        int numPages = computeNumberOfFolioPages();
        // numFolios computed from the foliation formula in Peter's catalog xml file
        // numPages computed from pages.xml
        boolean b = numFolios == numPages/2;
        if (!b) {
            b = false;
        }
        return b;
    }

    private void getNumber (BoundaryInfo bi, String s) {
        bi.actualValue = s;
        if (s.charAt(0) == '[') {
            bi.inBrackets = true;
        } else if (s.contains("[")) {
            bi.bracketedModifier = true;
        }

        s = s.replaceAll("[\\]\\[]", "");       // get rid of square brackets

        reNumber.reset(s);
        if (reNumber.find()) {
            bi.modifier = reNumber.getGroup(2);
            bi.value = Integer.parseInt(reNumber.getGroup(1));
        } else {
            reRoman.reset(s);
            if (reRoman.find()) {
                bi.isSmallRoman = true;
                bi.modifier = reRoman.getGroup(2);
                bi.value = convertRoman(reRoman.getGroup(1));
            } else {
                bi.error = true;
                Main.writeError(String.format("%s.xml: Can't match %s (in ManuscriptInfo.java)", catalogName, s));
            }
        }
    }

    private int convertRoman (String s) {
        int n = 0;
        int len = s.length();
        for (int i=0; i<len; i++) {
            char ch = s.charAt(i);
            switch (ch) {
                case 'i':
                    if (i == len-1) {
                        n += 1;
                    } else if (s.charAt(i+1) == 'i') {
                        n += 1;
                    } else {
                        n -= 1;
                    }
                    break;
                case 'v' :
                    n += 5;
                    break;
                case 'x' :
                    n += 10;
                    break;
            }
        }
        return n;
    }

// Will merge comments made on images that contain parts of two different texts.
    public void consolidateComments () {
        numComments = 0;
        // there will be no comments for notImaged but need to treat it like any other image so add it to comments.
        comments.put(notImaged, new ArrayList<String>());


        for (SectionInfo si : imageInfo) {
            if (si.textType.endsWith("Supp")) {
                continue;
            }
            for (FileInfo fi : si.files) {
                String name = fi.name;
                if (fi.name.equals(notImaged)) {
                    fi.textType = si.textType;
                    imageFiles.add(fi);     // there will be no comments on something that wasn't imaged but need to keep track of all image files. Each notImaged file is conceptually a different file in this program.
                } else {
                    ArrayList<String> a = comments.get(name);
                    if (a == null) {
                        a = new ArrayList<String>();
                        comments.put(name, a);
                        fi.textType = si.textType;
                        imageFiles.add(fi);
                    }
                    if (fi.anupamaComments.size() > 0) {
                        numComments++;
                        for (String ac : fi.anupamaComments) {
                            a.add(ac);
                        }
                    }
                }
            }
        }
    }

    public String generateIndexPageEntry () {
        String msg = "";
        if (numComments == 0) {
            msg = "<span style='color:red'>No Anupama comments</span>";
        } else {
            msg = String.format("%d Anupama comments", numComments);
        }
        if (foliationFormula == null) {
            return null;
        }
        if (imageFolioCountConsistent()) {
            return String.format("<a target='page' href='%s.html'>%s</a>  %s (#f=%d, #i=%d)&nbsp;&nbsp;&nbsp;%s<br/>\n",
                catalogName, catalogName, foliationFormula, numFolios, numImages, msg);
        } else {
            return String.format("<a target='page' href='%s.html'>%s</a>  %s <span style='color:red'>(#f=%d, #i=%d)</span>&nbsp;&nbsp;&nbsp;%s<br/>\n",
                catalogName, catalogName, foliationFormula, numFolios, numImages, msg);
        }
    }

    public String generateHTML (boolean croppingsOnly, boolean usePetite) {
        if (folioItems == null) {
            return null;
        }

        getPageIDs();

        if (directoryName.equals("Brown/32")) {
            directoryName = "Brown/32";
        }

        StringBuilder sb = new StringBuilder();
        generateHeadElement(sb, croppingsOnly, usePetite);
        int numPages = folioIDs.size();
        int k = 0;
        int imageCount = imageFiles.size();
        for (int i=0; i<imageCount; i++) {
            FileInfo fi = imageFiles.get(i);
            int numFolioPagesOnImage = fi.numFolioPages;
            sb.append(String.format(htmlRowHeader, i, i, getImage(i)));
            if (numFolioPagesOnImage > 0) {
                String delim = "";
                for (int j = 0; j < numFolioPagesOnImage; j++) {
                    String pageNumber = null;
                    if (k >= folioIDs.size()) {
                        pageNumber = "???";
                    } else {
                        pageNumber = folioIDs.get(k);
                    }
                    k++;
                    sb.append(String.format("%s%s", delim, pageNumber));
                    delim = "/";
                }
            } else {        // Image contains no folio images so it can't be mapped to folio IDs
            }

            sb.append(String.format(htmlRowAnupama, getAnupamaComments(i)));
        }
        sb.append(htmlFooter);
        return sb.toString();

    }

    void getPageIDs () {
        if (folioIDs != null) {
            return;
        }

        boolean dump = false;

        folioIDs = new ArrayList<String>();
        String item = null;

        if (folioItems == null) {
            return;
        }


        for (FolioItem fi : folioItems) {
            int start = fi.start.value;
            int end;
            boolean smallRoman = fi.start.isSmallRoman;
            boolean inBrackets = fi.start.inBrackets;
            boolean bracketedModifier = fi.start.bracketedModifier;
            String modifier = "";
            if (fi.end ==  null) {
                end = start;
                modifier = fi.start.modifier;
                if (bracketedModifier) {
                    modifier = String.format("[%s]", modifier);
                }
            } else {
                end = fi.end.value;
            }
            item = fi.start.actualValue;
            if (!item.endsWith("]")) {
                item += (fi.inBrackets ? "]" : "");
            }
            if (fi.enumerationID != null) {
                item = String.format("%s:%s", fi.enumerationID, item);
            }

            folioIDs.add(item + 'r');
            folioIDs.add(item + 'v');
            for (int i=start+1; i<=end-1; i++) {
                if (smallRoman) {
                    item = toRomanNumeral(i);
                } else {
                    item = Integer.toString(i);
                }

                if (inBrackets) {
                    item = String.format("[%s]", item);
                }

                if (fi.enumerationID != null) {
                    item = String.format("%s:%s", fi.enumerationID, item);
                }

                folioIDs.add(item + "r");
                folioIDs.add(item + "v");
            }
            if (end != start) {
                item = "";
                if (fi.end.actualValue.charAt(0) != '[') {
                    item = fi.inBrackets ? "[" : "";
                }
                item += fi.end.actualValue;
                if (fi.enumerationID != null) {
                    item = String.format("%s:%s", fi.enumerationID, item);
                }

                folioIDs.add(item + 'r');
                folioIDs.add(item + 'v');
            }
        }

        if (dump) {
            for (String pid : folioIDs) {
                System.out.println(pid);
            }
        }
    }

    private String getAnupamaComments (int n) {
        StringBuilder sb = new StringBuilder();
        if (n < imageFiles.size()) {
            ArrayList<String> a = comments.get(imageFiles.get(n).name);
            for (String c : a) {
                sb.append(String.format("<br/>%s", c));
            }
        } else {
            sb.append("<br/>???");
        }
        return sb.toString();
    }


    // Possibly adds a text type if this is first image of that text type
    private String getImage (int n) {
        String name;
        if (n < imageFiles.size()) {
            String textType = "";
            if (n > 0) {
                if (!imageFiles.get(n).textType.equals(imageFiles.get(n-1).textType)) {
                    textType = String.format("<b>%s</b><br/>", imageFiles.get(n).textType);
                }
            } else {
                textType = String.format("<b>%s</b><br/>", imageFiles.get(0).textType);
            }
            name = imageFiles.get(n).name;
            if (name.equals(notImaged)) {
                name = "<b>Not Imaged</b>";
            }
            return String.format("%s%s", textType, name);
        } else {
            return "???";
        }
    }

    private void generateImageArray(StringBuilder sb) {
        sb.append("\nvar im = new Array();\n");
        Hashtable<String, Boolean> h = new Hashtable<String, Boolean>();
        int n = 0;
        for (SectionInfo pi : imageInfo) {
            for (FileInfo fi : pi.files) {
                String page = fi.name;
                if (page.equals(notImaged)) {
                        sb.append(String.format("im[%d] = {img:'/%s', numPages:%d};\n",
                                n++, notImaged.substring(0, notImaged.length()-4), fi.numFolioPages));
                } else if (h.get(page) == null) {       // in case image is used in more than one pi
                    h.put(page, Boolean.TRUE);
                    reRemoveJPG.reset(page);
                    if (reRemoveJPG.find()) {
                        sb.append(String.format("im[%d] = {img:'/%s', numPages:%d};\n",
                                n++, page.substring(0, page.length()-4), fi.numFolioPages));
                    } else {
                        Main.writeError(String.format("%s: Unexpected page name: %s (in ManuscriptInfo.java)", directoryName, page));
                    }
                }
            }
        }
        System.out.format("%s has %d images\n", this.catalogName, n);
    }

    private void generateHeadElement (StringBuilder sb, boolean croppingOnly, boolean usePetite) {
        generateImageArray(sb);
        String actualName = null;
        String[] fields = directoryName.split("/");
        if (usePetite) {
            actualName = String.format("%s/%s", fields[0], fields[1]);
        } else {
            actualName = String.format("%s_s/%s", fields[0], fields[1]);
        }
        sb.append(String.format(scriptTemplate,
                croppingOnly ? "true" : "false",
                Main.dirOriginal,
                Main.dirCropped,
                actualName,
                usePetite ? "/split/": "/",
                notImaged.substring(0, notImaged.length()-4)));
        String head = String.format(htmlHeader, sb.toString());
        sb.setLength(0);
        sb.append(head);
    }


    private String toRomanNumeral (int n) {
        StringBuilder sb = new StringBuilder();
        while (n > 10) {
            sb.append("x");
            n -= 10;
        }
        if (n >= 5) {
            sb.append("v");
            n -= 5;
        }

        if (n == 4) {
            sb.append("iv");
        } else {
            for (int i=0; i<n; i++) {
                sb.append("i");
            }
        }
        return sb.toString();
    }


    @Override
    public String toString () {
        StringBuilder sb = new StringBuilder();
        sb.append(String.format("(%s) ", foliationFormula==null ? "None" : foliationFormula));
        for (SectionInfo pi : imageInfo) {
                sb.append(String.format("%s (%d) ",
                    pi.textType, pi.files.size()));
        }
        return sb.toString();
    }

    class InsertionPointCompare implements Comparator<InsertionPoint> {

        public int compare(InsertionPoint ip1, InsertionPoint ip2) {
            return ip1.getStartLine() - ip2.getStartLine();
        }

    }
}
