package folio;

import java.io.File;
import java.util.Hashtable;
import org.sanskritlibrary.ElementStack;
import org.sanskritlibrary.SAXParser;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class XMLCatalog implements SAXParserHandler {

    private Hashtable<String, InsertionPoint> locusTargets = new Hashtable<String, InsertionPoint>();
    private Locator locator;
    private String filename;
    private String shortname;
    private SAXParser saxParser;
    private StringBuilder buffer = new StringBuilder();
    private ElementStack stack = new ElementStack();
    private InsertionPoint currentInsertionPoint;
    boolean collectFormula = false;
    String currentFormula = null;

    public XMLCatalog() {
        saxParser = new SAXParser(this);
    }

    public void parse(String filename) throws Exception {
        this.filename = filename;
        buffer.setLength(0);
        currentFormula = null;
        collectFormula = false;
        locusTargets.clear();
        shortname = new File(filename).getName();
        shortname = shortname.substring(0, shortname.indexOf("."));
        try {
            String fn = String.format("%s%s", Main.dirXMLCatalog, filename);
            saxParser.parse(fn);
        } catch (Exception ex) {
            System.out.format("In XMLCatalog.java: %s (%s)\n", filename, ex.getMessage());
        }
    }

    public String getFormula() {
        return currentFormula;
    }

    public Hashtable<String, InsertionPoint> getInsertionPoints() {
        return locusTargets;
    }

    public void warning(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void error(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void fatalError(SAXParseException exception) throws SAXException {
        String msg = String.format("[XML:PMS]%s:%d.%d %s (in XMLCatalog.java)", filename, locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        Main.writeError(msg);
        throw new SAXException("Boom!");
    }

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    public void startDocument() throws SAXException {
    }

    public void endDocument() throws SAXException {
    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    public void endPrefixMapping(String prefix) throws SAXException {
    }

    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        if (shortname.equals("penn2639")) {
            shortname = "penn2639";
        }
        stack.pushElement(qName, atts);
        if (qName.equals("formula")) {
            if (stack.getElement(stack.topIndex() - 1).qName.equals("foliation")) {
                collectFormula = true;
            }
        } else if (qName.equals("facsimile")) {
            String id = atts.getValue("xml:id");
            // if id is null is this an error?
            if (id != null) {
                currentInsertionPoint = locusTargets.get(id);
                if (currentInsertionPoint != null) {
                    currentInsertionPoint.setStartLine(locator.getLineNumber());
                } else {
                    Main.writeError(String.format("[XML:PMS]%s: No locus references facsimile with xml:id='%s (in XMLCatalog.java)", shortname, id));
                }
            } else {
                Main.writeError(String.format("No xml:id attribute on facsimile element (in XMLCatalog.java)", shortname, id));
            }
        } else if (qName.equals("locus")) {
            String target = atts.getValue("target");
            if (target != null) {       // if it has a target attribute then it is the xml:id of a facsimile element which is included later in the document.
                InsertionPoint ip = new InsertionPoint();
                ip.setID(target.substring(1));    // skip the #. There should be a facsimile element with this id
                ip.setFrom(atts.getValue("from"));
                ip.setTo(atts.getValue("to"));
                if (ip.getFrom() == null || ip.getTo() == null) {
                    Main.writeError(String.format("[XML:PMS]%s: possible problem at %d, from/to attribute is missing. (in XMLCatalog.java)", shortname, locator.getLineNumber()));
                }
                if (locusTargets.get(ip.getID()) != null) {
                    // Main.writeError(String.format("%s: possible problem at %d, '#%s' used more than once. (in XMLCatalog.java)", shortname, locator.getLineNumber(), ip.id));
                }
                locusTargets.put(ip.getID(), ip);
            }
        }
    }

    public void endElement(String uri, String localName, String qName) throws SAXException {
        if (qName.equals("formula") && collectFormula) {
            if (currentFormula == null) {           // assume that the first one is the one being used
                currentFormula = buffer.toString();
                buffer.setLength(0);
                collectFormula = false;
            }
        } else if (qName.equals("facsimile")) {
            if (currentInsertionPoint != null) {
                currentInsertionPoint.setEndLine(locator.getLineNumber());
                currentInsertionPoint = null;
            }
        }
        stack.popElement();
    }

    public void characters(char[] ch, int start, int length) throws SAXException {
        if (collectFormula) {
            buffer.append(new String(ch, start, length));
        }
    }
}
